/*------------------------------------------------------------------------------
;
;   Easy Loader.
;
;   Compatible with The Cowboy Boot Loader for Enhanced PIC12/PIC16,
;                   The Wellington Boot Loader for PIC18 and
;                   The Oxford Boot Loader for dsPIC/PIC24.
;
;   Copyright (c) 2014 - 2017 Darron M Broad
;   Copyright (c) 2014 Gerhard Bertelsmann
;
;-------------------------------------------------------------------------------

;-------------------------------------------------------------------------------
;   This file is part of Easy Loader.
;
;   Easy Loader is free software: you can redistribute it and/or
;   modify it under the terms of the GNU General Public License as published
;   by the Free Software Foundation.
;
;   Easy Loader is distributed in the hope that it will be
;   useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
;   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
;   GNU General Public License for more details.
;
;   You should have received a copy of the GNU General Public License along
;   with Easy Loader. If not, see http://www.gnu.org/licenses/
;-----------------------------------------------------------------------------*/

#include "easy-loader.h"

#undef DEBUG

/******************************************************************************
 *
 * Session
 *
 *****************************************************************************/

extern session_t easy;

/*******************************************************************************
 *
 * Send command to and get response from boot loader
 *
 ******************************************************************************/
int
doCommand(int outlen, int inlen)
{
#ifdef DEBUG
	program_t *program = (program_t *)&easy.buffer[0]; /* COMMAND */
	printf("\n");
	printf("%s() outlen=%d inlen=%d\n", __func__, outlen, inlen);
	printf("%s() command  = 0x%02X\n", __func__, program->command);
	printf("%s() address  = 0x%02X%02X%02X\n", __func__, program->addru, program->addrh, program->addrl);
	printf("%s() datasize = %d\n", __func__, program->datasize);
	int j = 0;
	for (int i = 0; i < outlen; ++i) {
		printf(" 0x%02X", easy.buffer[i]);
		if (++j == 16) {
			j = 0;
			printf("\n");
		}
	}
	if (j)
		printf("\n");
	j = 0;
	for (int i = 0; i < outlen; ++i) {
		printf(" %-4d", easy.buffer[i]);
		if (++j == 16) {
			j = 0;
			printf("\n");
		}
	}
	if (j)
		printf("\n");
#endif
	/*
	 * Send command
	 */

	assert(outlen > 0 && outlen < BUFLEN);

	int rc = fdio(outlen, OUT, TIMEOUT);
	if (rc != outlen) {
		if (rc < 0) {
			fprintf(stderr, "I/O error in write [%s].\n", strerror(errno));
			return -1;
		}
		fprintf(stderr, "I/O timed-out in write.\n");
		return -1;
	}

	/*
	 * Get any response
	 */

	if (inlen) {
		assert(inlen < BUFLEN);

		rc = fdio(inlen, IN, TIMEOUT);
		if (rc != inlen) {
			if (rc < 0) {
				fprintf(stderr, "I/O error in read [%s].\n", strerror(errno));
				return -1;
			}
			fprintf(stderr, "I/O timed-out in read.\n");
			return -1;
		}

		char c = easy.buffer[inlen - 1];
#ifdef DEBUG
		printf("RESPONSE = 0x%02X\n", c);
#endif
		if (c != RESPONSE_OK) {
			if (c == 'E')
				fprintf(stderr, "BOOT LOADER COMMAND ERROR.\n");
			else if (c == 'S')
				fprintf(stderr, "BOOT LOADER CHECKSUM ERROR.\n");
			else
				fprintf(stderr, "BOOT LOADER PROTOCOL ERROR [0x%02X != '%c'].\n",
					c & 0xFF, RESPONSE_OK);
			return -1;
		}
	}

	return 0;
}

/*******************************************************************************
 *
 * Do HELLO (determine version before hand)
 *
 ******************************************************************************/
int
helloCommand(void)
{
	bzero(easy.buffer, BUFLEN);

	easy.buffer[0] = COMMAND_HELLO;

	if (doCommand(1, sizeof(hello_t)) < 0)
		return -1;

	hello_t *hello = (hello_t *)&easy.buffer[0];

	easy.arch      = hello->wordsize;
	easy.writesize = hello->writesize;
	easy.erasesize = hello->erasesizeh << 8 | hello->erasesizel;
	easy.startaddr = hello->startu << 16 | hello->starth << 8 | hello->startl;
	easy.eesize    = hello->eesize * 64;
	easy.option    = hello->option;

	fprintf(stdout, "BOOT LOADER VERSION = %d\n", easy.version);
	fprintf(stdout, "BOOT LOADER OPTIONS = %d\n", easy.option);
	fprintf(stdout, "BOOT LOADER ADDRESS = 0x%06X\n", easy.startaddr);
	fprintf(stdout, "WORD SIZE   = %d\n", easy.arch);
	fprintf(stdout, "ERASE SIZE  = %d\n", easy.erasesize);
	fprintf(stdout, "WRITE SIZE  = %d\n", easy.writesize);
	fprintf(stdout, "EEPROM SIZE = %d\n", easy.eesize);
	
	return 0;
}

/*******************************************************************************
 *
 * Do RUN
 *
 ******************************************************************************/
int
runCommand(void)
{
	bzero(easy.buffer, BUFLEN);

	easy.buffer[0] = COMMAND_RUN;

	return doCommand(1, 0);
}

/*******************************************************************************
 *
 * Do Version
 *
 ******************************************************************************/
int
versionCommand(void)
{
	bzero(easy.buffer, BUFLEN);

	easy.buffer[0] = COMMAND_VERSION;

	if (doCommand(1, 2) < 0)
		return -1;

	easy.version = easy.buffer[0];

	return 0;
}

/*******************************************************************************
 *
 * Set programming checksum
 *
 ******************************************************************************/
void
programChecksum(uint16_t ix)
{
	uint8_t sum = 0;
	uint16_t i, ofs = PROGRAM_SIZE + ix - 1;

	for (i = 0; i < ofs; ++i)
		sum -= easy.buffer[i];

	easy.buffer[i] = sum;
}
