/*------------------------------------------------------------------------------
;
;   Easy Loader.
;
;   Compatible with The Cowboy Boot Loader for Enhanced PIC12/PIC16,
;                   The Wellington Boot Loader for PIC18 and
;                   The Oxford Boot Loader for dsPIC/PIC24.
;
;   Copyright (c) 2014 - 2017 Darron M Broad
;   Copyright (c) 2014 Gerhard Bertelsmann
;
;-------------------------------------------------------------------------------

;-------------------------------------------------------------------------------
;   This file is part of Easy Loader.
;
;   Easy Loader is free software: you can redistribute it and/or
;   modify it under the terms of the GNU General Public License as published
;   by the Free Software Foundation.
;
;   Easy Loader is distributed in the hope that it will be
;   useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
;   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
;   GNU General Public License for more details.
;
;   You should have received a copy of the GNU General Public License along
;   with Easy Loader. If not, see http://www.gnu.org/licenses/
;-----------------------------------------------------------------------------*/

#include "VERSION"

#include "easy-loader.h"

/*******************************************************************************
 *
 * Easy loader session
 *
 ******************************************************************************/

session_t easy = {0};

/*******************************************************************************
 *
 * HOWTO
 *
 ******************************************************************************/
void
usage(const char *msg, int err)
{
	fprintf(stderr, "USAGE: easy-loader [OPTIONS] DEVICE [HEX-FILE]\n\n");

	if (msg)
		fprintf(stderr, "%s\n\n", msg);

	fprintf(stderr, "Options:\n"
		" -a     arduino (DTR + CAP) reset\n"
		" -b N   use TTY baud rate N\n"
		" -d     hang up DTR after serial port open\n"
		" -e     read EEPROM\n"
		" -f     read flash\n"
		" -h     HELLO only\n"
		" -i N   use CAN bus message id N\n"
		" -k N   send BRK for N ms after serial port open\n"
		" -l H|R log serial input when done (H = hex, R = raw)\n"
		" -n     do not run firmware on exit\n"
		" -p     run pickle pctrl to reset device\n"
		" -q     quiet erase/write\n"
		" -r     hang up RTS after serial port open\n"
		" -s     simulate erase/write\n"
		" -t     time to wait in ms after serial port open\n"
		" -v     verify\n"
		" -w N   winklepicker mode N (1 or 2) reset\n"
		" -x     drain serial output after each transmitted byte\n"
		" -z A   erase page at address A\n"
		"\n");

	fprintf(stderr, "Version: %d\n", VERSION);

	exit(err);
}

/*******************************************************************************
 *
 * Console log
 *
 ******************************************************************************/
static void
hexconsole(void)
{
	int rc;
	char s[WIDTH + 1] = {0};
	int i = 0;

	while (1) {
		rc = fdio(1, 1, INT_MAX);
		if (rc < 0)
			return;
		if (rc == 0)
			continue;

		fprintf(stdout, "%02X ", easy.buffer[0]);
		fflush(stdout);

		if (easy.buffer[0] >= ' ' && easy.buffer[0] < 127)
			s[i++] = easy.buffer[0];
		else
			s[i++] = '.';

		if (i == WIDTH) {
			printf("%s\n", s);
			fflush(stdout);
			i = 0;
		}
	}
}

static void
rawconsole(void)
{
	int rc;

	while (1) {
		rc = fdio(1, 1, INT_MAX);
		if (rc < 0)
			return;
		if (rc == 0)
			continue;

		putc(easy.buffer[0], stdout);
		fflush(stdout);
	}
}

/*******************************************************************************
 *
 * Take it easy.
 *
 ******************************************************************************/
int
main(int argc, char **argv)
{
	int opt, nargs = 2, eeprom_read = 0, flash_read = 0, pctrl = 0, hello_only = 0, zap = -1, log = 0;

	easy.info = stdout;
	easy.cid = CAN_ID;
	easy.delay = 10000; /* 20ms */
	easy.autorun = 1;
	easy.baudrate = 115200;

	opterr = 0;
	while ((opt = getopt(argc, argv, "ab:defhi:k:l:nrpqst:vw:xz:")) != -1) {
		switch (opt) {
		case 'a':
			easy.wiener = 10000;
			break;
		case 'b':
			easy.baudrate = strtoul(optarg, NULL, 0);
			break;
		case 'd':
			easy.dtr = 10000; /* 10ms */
			break;
		case 'e':
			eeprom_read = 1;
			nargs = 1;
			break;
		case 'f':
			flash_read = 1;
			nargs = 1;
			break;
		case 'h':
			hello_only = 1;
			nargs = 1;
			break;
		case 'i':
			easy.cid = strtoul(optarg, NULL, 0);
			break;	
		case 'k':
			easy.brk = strtoul(optarg, NULL, 0) * 1000;
			break;
		case 'l':
			log = tolower(optarg[0]);
			if (log != 'h' && log != 'r')
				usage("Invalid arg.", EX_USAGE);
			break;
		case 'n':
			easy.autorun = 0;
			break;
		case 'p':
			pctrl = 1;
			break;
		case 'q':
			easy.info = NULL;
			break;
		case 'r':
			easy.rts = 10000; /* 10ms */
			break;
		case 's':
			easy.simulate = 1;
			break;
		case 't':
			easy.delay = strtoul(optarg, NULL, 0) * 1000;
			break;
		case 'v':
			easy.verify = 1;
			break;
		case 'w':
			easy.winkle = strtoul(optarg, NULL, 0);
			if (easy.winkle < 1 || easy.winkle > 2)
				usage("Invalid arg.", EX_USAGE);
			break;
		case 'x':
			easy.txdrain = 1;
			break;
		case 'z':
			zap = strtoul(optarg, NULL, 0);
			nargs = 1;
			break;
		default:
			usage("Unknown option.", EX_USAGE);
		}
	}
	argc -= optind;
	argv += optind;
	if (argc != nargs && log == 0)
		usage("Invalid args.", EX_USAGE);

	if (fdopendev(argv[0]) < 0) {
		fprintf(stderr, "I/O device open failed [%s].\n", argv[0]);
		exit(EX_OSERR);
        }

	if (pctrl && (run("pctrl restore") < 0)) {
		fprintf(stderr, "I/O failed [PCTRL].\n");
		fdclosedev();
		exit(EX_SOFTWARE);
	}

	if (versionCommand() < 0) {
		fprintf(stderr, "I/O failed [VERSION].\n");
		fdclosedev();
		exit(EX_SOFTWARE);
	}

	if (easy.version != VERSION) {
		fprintf(stderr, "BOOT LOADER UNSUPPORTED.\n");
		fdclosedev();
		exit(EX_SOFTWARE);
	}

	if (helloCommand() < 0) {
		fprintf(stderr, "I/O failed [HELLO].\n");
		fdclosedev();
		exit(EX_SOFTWARE);
	}

	if (!hello_only) {
		if (zap >= 0 && (pic_zap(zap) < 0)) {
			fprintf(stderr, "I/O failed [FLASH_ERASE].\n");
			fdclosedev();
			exit(EX_SOFTWARE);
		}
		if (flash_read && (pic_dumpFlash() < 0)) {
			fprintf(stderr, "I/O failed [FLASH_READ].\n");
			fdclosedev();
			exit(EX_SOFTWARE);
		}
		if (eeprom_read && (pic_dumpEEPROM() < 0)) {
			fprintf(stderr, "I/O failed [EE_READ].\n");
			fdclosedev();
			exit(EX_SOFTWARE);
		}
		if (argv[1] && (pic_program(argv[1]) < 0)) {
			fprintf(stderr, "I/O failed [FLASH/EE_WRITE].\n");
			fdclosedev();
			exit(EX_SOFTWARE);
		}
	}

	if (easy.autorun && (runCommand() < 0)) {
		fprintf(stderr, "I/O failed [RUN].\n");
		fdclosedev();
		exit(EX_SOFTWARE);
	}

	if (log == 'h')
		hexconsole();
	else
		rawconsole();

	fdclosedev();
	exit(EX_OK);
}
