/*------------------------------------------------------------------------------
;
;   Easy Loader.
;
;   Compatible with The Cowboy Boot Loader for Enhanced PIC12/PIC16,
;                   The Wellington Boot Loader for PIC18 and
;                   The Oxford Boot Loader for dsPIC/PIC24.
;
;   Copyright (c) 2014 - 2017 Darron M Broad
;   Copyright (c) 2014 Gerhard Bertelsmann
;
;-------------------------------------------------------------------------------

;-------------------------------------------------------------------------------
;   This file is part of Easy Loader.
;
;   Easy Loader is free software: you can redistribute it and/or
;   modify it under the terms of the GNU General Public License as published
;   by the Free Software Foundation.
;
;   Easy Loader is distributed in the hope that it will be
;   useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
;   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
;   GNU General Public License for more details.
;
;   You should have received a copy of the GNU General Public License along
;   with Easy Loader. If not, see http://www.gnu.org/licenses/
;-----------------------------------------------------------------------------*/

#include "easy-loader.h"

#undef DEBUG

/******************************************************************************
 *
 * Session
 *
 *****************************************************************************/

extern session_t easy;

/*******************************************************************************
 *
 * Read Hex Nibble
 *
 ******************************************************************************/
static uint8_t
hex2nib(char c)
{
	uint8_t n = 0;

	if (c >= 'a' && c <= 'f') {
		n = c - 'a' + 10;
	}
	else if (c >= 'A' && c <= 'F') {
		n = c - 'A' + 10;
	}
	else if (c >= '0' && c <= '9') {
		n = c - '0';
	}
	return n;
}

/*******************************************************************************
 *
 * Read Hex Byte
 *
 ******************************************************************************/
static inline uint8_t
hex2byt(char *s)
{
	return hex2nib(s[0]) << 4 | hex2nib(s[1]);
}

/*******************************************************************************
 *
 * Dump Hex
 *
 ******************************************************************************/
static inline void
dump_hex(uint32_t a, uint8_t *s, int n)
{
	int i = 0;

	printf("[%08X] ", a);
	while (i < n) {
		printf("0x%02X ", s[i]);
		i++;
	}
	printf("\n");
}

/*******************************************************************************
 *
 * Read Hex Line
 *
 ******************************************************************************/
static int
gethex(FILE *fp, uint16_t *addr, uint8_t *typ, uint8_t *bin, uint8_t *binlen)
{
	char line[BUFLEN];
	int i;
	uint8_t sum, len;

	if (fgets(line, BUFLEN, fp) == NULL)
		return 0;

	if (line[0] != ':')
		return -1;

	line[BUFMAX] = '\0';

	i = strlen(line) - 1;
	while (i >= 0 && (line[i] == '\r' || line[i] == '\n'))
		line[i--] = '\0';
		
	if (i & 1 || i < 10)
		return -1;

	sum = 0;
	*binlen = 0;
	for (i = 1; line[i]; i += 2) {
		bin[*binlen] = hex2byt(&line[i]);
		sum += bin[(*binlen)++];
	}
	if (sum)
		return -1;

	len = bin[0];
	if (*binlen != (len + 5))
		return -1;

	*addr = bin[1] << 8 | bin[2];
	*typ = bin[3];
	
	return len;
}

/*******************************************************************************
 *
 * Read Hex File
 *
 * NB. 16-bit arrays hold 8-bit data so 0xFFFF can indicate an unused element.
 *
 ******************************************************************************/
int
readFile(const char *file, uint16_t *flash, uint16_t *eeprom, uint32_t *nflash, uint32_t *neeprom)
{
	FILE *fp;
	uint16_t addr = 0, ext_addr = 0;
	uint8_t typ = 0, bin[BUFLEN] = {0}, binlen = 0, *data = &bin[4];
	uint32_t i, j, address;
	uint32_t flashhigh, eeaddr, eehigh;

	*nflash = 0;
	*neeprom = 0;

	fp = fopen(file, "rb");
	if (!fp) {
		return -1;
	}

	if (easy.arch == 14) {
		flashhigh = easy.startaddr << 1;
		eeaddr = 0x0001E000;
		eehigh = 0x0001E000 + (easy.eesize << 1);
	} else if (easy.arch == 16) {
		flashhigh = easy.startaddr;
		eeaddr = 0x00F00000;
		eehigh = 0x00F00000 + easy.eesize;
	} else if (easy.arch == 24) {
		flashhigh = easy.startaddr << 1;
		eeaddr = 0x01000000 - (easy.eesize << 1);
		eehigh = 0x01000000;
	} else {
		fprintf(stderr, "ARCH UNKNOWN.\n");
		exit(EX_SOFTWARE);
	}

	while (typ != 1 && (gethex(fp, &addr, &typ, bin, &binlen) > 0)) {
		switch (typ) {
		case 0:	/* data */
			binlen -= 5;
			address = ext_addr << 16 | addr;
#ifdef DEBUG
			dump_hex(address, data, binlen);
#endif
			for (i = 0; i < binlen; ++i) {
				j = address + i;
				if (j < flashhigh) {
					flash[j] = data[i];
					(*nflash)++;
				}
				else if (eeprom && (j >= eeaddr && j < eehigh)) {
					eeprom[j - eeaddr] = data[i];
					(*neeprom)++;
				}
			}
			break;
		case 4: /* address */
			ext_addr = bin[4] << 8 | bin[5];
			break;
		}
	}

	fclose(fp);

	return 0;
}
