/*------------------------------------------------------------------------------
;
;   Easy Loader.
;
;   Compatible with The Cowboy Boot Loader for Enhanced PIC12/PIC16,
;                   The Wellington Boot Loader for PIC18 and
;                   The Oxford Boot Loader for dsPIC/PIC24.
;
;   Copyright (c) 2014 - 2017 Darron M Broad
;   Copyright (c) 2014 Gerhard Bertelsmann
;
;-------------------------------------------------------------------------------

;-------------------------------------------------------------------------------
;   This file is part of Easy Loader.
;
;   Easy Loader is free software: you can redistribute it and/or
;   modify it under the terms of the GNU General Public License as published
;   by the Free Software Foundation.
;
;   Easy Loader is distributed in the hope that it will be
;   useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
;   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
;   GNU General Public License for more details.
;
;   You should have received a copy of the GNU General Public License along
;   with Easy Loader. If not, see http://www.gnu.org/licenses/
;-----------------------------------------------------------------------------*/

#include "easy-loader.h"

/******************************************************************************
 *
 * Session
 *
 *****************************************************************************/

extern session_t easy;

/*******************************************************************************
 *
 * Select in or out
 *
 ******************************************************************************/
static int
fdselect(int io, int timeout)
{
	int rc = 0;
	struct timeval tv;
	fd_set fdset;

	tv.tv_sec = timeout;
	tv.tv_usec = 0;

	FD_ZERO(&fdset);
	FD_SET(easy.fd, &fdset);

	if (io == IN)
		rc = select(easy.fd + 1, &fdset, NULL, NULL, &tv);
	else
		rc = select(easy.fd + 1, NULL, &fdset, NULL, &tv);

	if (rc < 0)
		return rc;

	if (!FD_ISSET(easy.fd, &fdset))
		return 0; /* Timed-out */

	if (rc != 1)
		return 0; /* Timed-out */

	return 1;
}

/*******************************************************************************
 *
 * Read in or write out
 *
 ******************************************************************************/
static int
fdreadwrite(uint8_t *buffer, int buflen, int io)
{
	int rc = 0;

	if (io == IN) {
		if (easy.fdtyp == CANBUS) {
			rc = canRead(buffer, buflen);
		} else if (easy.fdtyp == UART) {
			rc = read(easy.fd, buffer, buflen);
		}
	} else {
		if (easy.fdtyp == CANBUS) {
			rc = canWrite(buffer, buflen);
		} else if (easy.fdtyp == UART) {
			if (easy.txdrain) {
				rc = write(easy.fd, buffer, 1);
				if (rc == 1)
					tcdrain(easy.fd);
			} else {
				rc = write(easy.fd, buffer, buflen);
			}
		}
	}

	return rc;
}

/*******************************************************************************
 *
 * I/O
 *
 ******************************************************************************/
int
fdio(int buflen, int io, int timeout)
{
	int rc = 0, nb = 0;

	while (nb < buflen) {
		rc = fdselect(io, timeout);
		if (rc < 0) {
			if (errno == EINTR)
				continue;
			fprintf(stderr, "Select failed [%d/%d/%d].\n", io, nb, buflen);
			return rc;
		}
		if (rc == 0) {
			fprintf(stderr, "Select timed-out [%d/%d/%d]\n", io, nb, buflen);
			return 0;
		}
		rc = fdreadwrite(&easy.buffer[nb], buflen - nb, io);
		if (rc < 0) {
			if (errno == EINTR || errno == EAGAIN)
				continue;
			fprintf(stderr, "R/W failed [%d/%d/%d].\n", io, nb, buflen);
			return rc;
		}
		if (rc == 0) {
			fprintf(stderr, "EOF [%d/%d/%d].\n", io, nb, buflen);
			break;
		}
		nb += rc;
	}

	return nb;
}

/*******************************************************************************
 *
 * Open
 *
 ******************************************************************************/
int
fdopendev(char *dev)
{
	if (dev[0] == '/') {
		easy.fdtyp = UART;
		easy.fd = openDevice(dev);
	} else {
		easy.fdtyp = CANBUS;
		easy.fd = openCanSock(dev);
	}

	return easy.fd;
}

/*******************************************************************************
 *
 * Close
 *
 ******************************************************************************/
void
fdclosedev(void)
{
	if (easy.fdtyp == UART)
		closeDevice();

	close(easy.fd);
}
