/*------------------------------------------------------------------------------
;
;   Easy Loader.
;
;   Compatible with The Cowboy Boot Loader for Enhanced PIC12/PIC16,
;                   The Wellington Boot Loader for PIC18 and
;                   The Oxford Boot Loader for dsPIC/PIC24.
;
;   Copyright (c) 2014 - 2017 Darron M Broad
;   Copyright (c) 2014 Gerhard Bertelsmann
;
;-------------------------------------------------------------------------------

;-------------------------------------------------------------------------------
;   This file is part of Easy Loader.
;
;   Easy Loader is free software: you can redistribute it and/or
;   modify it under the terms of the GNU General Public License as published
;   by the Free Software Foundation.
;
;   Easy Loader is distributed in the hope that it will be
;   useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
;   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
;   GNU General Public License for more details.
;
;   You should have received a copy of the GNU General Public License along
;   with Easy Loader. If not, see http://www.gnu.org/licenses/
;-----------------------------------------------------------------------------*/

#include "easy-loader.h"

/******************************************************************************
 *
 * Session
 *
 *****************************************************************************/

extern session_t easy;

/*******************************************************************************
 *
 * Fix GOTO
 *
 *  Preserve boot loader reset vector and store application reset vector in
 *  bootloader GOTO application vector.
 *
 ******************************************************************************/
static void
pic14_fixGOTO(uint16_t *flash)
{
	if (!pic_isBlank16(flash, easy.erasesize)) {
		uint32_t gotoapp = (easy.startaddr << 1) - 12;

		/* MOVLB 0  */
		flash[gotoapp + 1] = 0x01;
		flash[gotoapp + 0] = 0x40;
		/* MOVLP 0 */
		flash[gotoapp + 3] = 0x31;
		flash[gotoapp + 2] = 0x80;
		/* GOTO application */
		flash[gotoapp + 4]  = flash[0] & 0xFF;	/* 0 */
		flash[gotoapp + 5]  = flash[1] & 0xFF;
		flash[gotoapp + 6]  = flash[2] & 0xFF;	/* 1 */
		flash[gotoapp + 7]  = flash[3] & 0xFF;
		flash[gotoapp + 8]  = flash[4] & 0xFF;	/* 2 */
		flash[gotoapp + 9]  = flash[5] & 0xFF;
		flash[gotoapp + 10] = flash[6] & 0xFF;	/* 3 */
		flash[gotoapp + 11] = flash[7] & 0xFF;

		/* MOVLB 0 */
		flash[1] = 0x01;
		flash[0] = 0x40;
		/* MOVLP HIGH bootloader */
		flash[3] = 0x31;
		flash[2] = 0x80 | ((easy.startaddr >> 8) & 0x7F);
		/* GOTO bootloader */
		flash[5] = 0x28 | ((easy.startaddr & 0x700) >> 8);
		flash[4] = easy.startaddr & 0xFF;
		/* NOP */
		flash[6] = 0;
		flash[7] = 0;
	}
}

/*******************************************************************************
 *
 * Memory copy & compare
 *
 ******************************************************************************/
static void
pic14_memcpy(uint8_t *dst, uint16_t *src)
{
	for (uint32_t i = 0; i < (easy.writesize << 1); i += 2) {
		dst[i + 0] = src[i + 1];
		dst[i + 1] = src[i + 0];
	}
}

static int
pic14_memcmp(uint16_t *s1, uint8_t *s2, uint32_t n)
{
	uint32_t i, wordout, wordin;

	for (i = 0; i < n; i += 2) {
		wordin = (s2[i]) | (s2[i + 1] << 8);
		wordout = (s1[i] & 0xFF) | ((s1[i + 1] & 0x3F) << 8);
		if (wordin != wordout)
			return -1;
	}

	return 0;
}

/*******************************************************************************
 *
 * Erase row
 *
 ******************************************************************************/
int
pic14_erase_row(uint32_t raddress)
{
	program_t *program = (program_t *)&easy.buffer[0]; /* COMMAND */

	/* Erase row */
	program->command = COMMAND_FLASH_ERASE;
	program->addru = (raddress & 0xFF0000) >> 16;
	program->addrh = (raddress & 0x00FF00) >> 8;
	program->addrl = (raddress & 0x0000FF);
	program->datasize = 1; /* CHECKSUM */
	programChecksum(0);
	if (doCommand(PROGRAM_SIZE, 1) < 0)
		return -1;

	return 0;
}

/*******************************************************************************
 *
 * Verify row
 *
 ******************************************************************************/
static int
pic14_verify_row(uint16_t *flash, uint32_t address)
{
	program_t *program = (program_t *)&easy.buffer[0]; /* COMMAND */
	uint32_t faddress = (address << 1);

	program->command = COMMAND_FLASH_READ;
	program->addru = (address & 0xFF0000) >> 16;
	program->addrh = (address & 0x00FF00) >> 8;
	program->addrl = (address & 0x0000FF);
	program->datasize = 1; /* CHECKSUM */
	programChecksum(0);
	if (doCommand(PROGRAM_SIZE, 1 + PIC14_WORDS2BYTES(easy.writesize)) < 0)
		return -1;
	if (pic14_memcmp(&flash[faddress], easy.buffer, PIC14_WORDS2BYTES(easy.writesize)) == 0)
		fprintf(stdout, " VERIFY OK\n");
	else
		fprintf(stderr, " VERIFY ERROR\n");

	return 0;
}

/*******************************************************************************
 *
 * Upload to flash
 *
 *  Erase and Write rows. For some devices the erase row is larger than
 *  the write row.
 *
 ******************************************************************************/
static int
pic14_uploadFlash(uint16_t *flash)
{
	program_t *program = (program_t *)&easy.buffer[0]; /* COMMAND */

	/*
	 * For each erase row
	 */

	for (uint32_t eaddress = 0; eaddress < easy.startaddr; eaddress += easy.erasesize) {
		uint32_t faddress = (eaddress << 1);
		assert(faddress < (easy.startaddr << 1));
		if (pic_isBlank16(&flash[faddress], (easy.erasesize << 1)))
			continue;

		/*
		 * First write row
		 */

		if (pic_isBlank16(&flash[faddress], (easy.writesize << 1))) {

			/*
			 * Erase only
			 *
			 * This operation will occur when the first write row of an erase block
			 * is empty.
			 *
			 * When erase size == row size, this can only happen when a block is
			 * explicitly empty (all bytes -1) in the hex file
			 */

			/* Simulate? */
			if (!easy.simulate) {
				/* Erase row */
				if (pic14_erase_row(eaddress) < 0)
					return -1;
			}
			/* Information */
			if (easy.info)
				fprintf(easy.info, "ERASE FLASH 0x%06X\n", eaddress);
		} else {

			/*
			 * Erase then write
			 *
			 * When erase size == row size, this is the only operation performed
			 * when data is written
			 */

			/* Simulate? */
			if (!easy.simulate) {
				/* Erase then write row */
				program->command = COMMAND_FLASH_ERASE;
				program->addru = (eaddress & 0xFF0000) >> 16;
				program->addrh = (eaddress & 0x00FF00) >> 8;
				program->addrl = (eaddress & 0x0000FF);
				program->datasize = 1 + easy.writesize; /* CHECKSUM + data */
				pic14_memcpy(program->data, &flash[faddress]);
				programChecksum(PIC14_WORDS2BYTES(easy.writesize));
				if (doCommand(PROGRAM_SIZE + PIC14_WORDS2BYTES(easy.writesize), 1) < 0)
					return -1;
			}
			/* Information */
			if (easy.info)
				fprintf(easy.info, "ERASE FLASH 0x%06X %4d BYTES\n",
					eaddress, PIC14_WORDS2BYTES(easy.writesize));
			/* Verify */
			if (easy.verify) {
				if (pic14_verify_row(flash, eaddress) < 0)
					return -1;
			}
		}

		/*
		 * Next write row(s)
		 *
		 * This step is performed only when erase size > row size
		 */

		for (uint32_t i = easy.writesize; i < easy.erasesize; i += easy.writesize) {
			uint32_t raddress = eaddress + i;
			faddress = (raddress << 1);
			assert(faddress < (easy.startaddr << 1));
			if (pic_isBlank16(&flash[faddress], (easy.writesize << 1)))
				continue;

			/*
			 * Write
			 */

			/* Simulate? */
			if (!easy.simulate) {
				/* Write row */
				program->command = COMMAND_FLASH_WRITE;
				program->addru = (raddress & 0xFF0000) >> 16;
				program->addrh = (raddress & 0x00FF00) >> 8;
				program->addrl = (raddress & 0x0000FF);
				program->datasize = 1 + easy.writesize; /* CHECKSUM + data */
				pic14_memcpy(program->data, &flash[faddress]);
				programChecksum(PIC14_WORDS2BYTES(easy.writesize));
				if (doCommand(PROGRAM_SIZE + PIC14_WORDS2BYTES(easy.writesize), 1) < 0)
					return -1;
			}
			/* Information */
			if (easy.info)
				fprintf(easy.info, "WRITE FLASH 0x%06X %4d BYTES\n",
					raddress, PIC14_WORDS2BYTES(easy.writesize));
			/* Verify */
			if (easy.verify) {
				if (pic14_verify_row(flash, raddress) < 0)
					return -1;
			}
		}
	}

	return 0;
}

/*******************************************************************************
 *
 * Upload to EEPROM
 *
 ******************************************************************************/
static int
pic14_uploadEEPROM(uint16_t *eeprom)
{
	program_t *program = (program_t *)&easy.buffer[0]; /* COMMAND */

	for (uint32_t i = 0; i < (easy.eesize << 1); i += 2) {
		uint32_t eaddress = i >> 1;

		if (eeprom[i] == 0xFFFF)
			continue;

		/* Simulate? */
		if (!easy.simulate) {
			/* Write byte */
			program->command = COMMAND_EE_WRITE;
			program->addru = 0;
			program->addrh = (eaddress & 0x00FF00) >> 8;
			program->addrl = (eaddress & 0x0000FF);
			program->datasize = 2; /* CHECKSUM + data */
			program->data[0] = eeprom[i];	/* Low  byte */
			program->data[1] = 0;		/* High bits */
			programChecksum(PIC14_WORDS2BYTES(1));
			if (doCommand(PROGRAM_SIZE + PIC14_WORDS2BYTES(1), 1) < 0)
				return -1;
		}
		/* Information */
		if (easy.info)
			fprintf(easy.info, "WRITE EEPROM 0x%04X = 0x%02X\n",
				eaddress, eeprom[i]);
		/* Verify */
		if (easy.verify) {
			program->command = COMMAND_EE_READ;
			program->addru = 0;
			program->addrh = (eaddress & 0x00FF00) >> 8;
			program->addrl = (eaddress & 0x0000FF);
			program->datasize = 1; /* CHECKSUM */
			programChecksum(0);
			if (doCommand(PROGRAM_SIZE, 2) < 0)
				return -1;
			if (easy.buffer[0] == eeprom[i]) {
				fprintf(stdout, " VERIFY OK\n");
			} else
				fprintf(stderr, " VERIFY ERROR\n");
		}
	}

	return 0;
}

/*******************************************************************************
 *
 * Dump device flash
 *
 ******************************************************************************/
int
pic14_dumpFlash(void)
{
	int rc = 0, c = 0;
	program_t *program = (program_t *)&easy.buffer[0];
	uint16_t *flash = xmalloc(easy.startaddr * sizeof(uint16_t), -1);
	uint32_t ix = 0;

	for (uint32_t address = 0; address < easy.startaddr; address += easy.writesize) {
		program->command = COMMAND_FLASH_READ;
		program->addru = (address & 0xFF0000) >> 16;
		program->addrh = (address & 0x00FF00) >> 8;
		program->addrl = (address & 0x0000FF);
		program->datasize = 1; /* CHECKSUM */
		programChecksum(0);
		rc = doCommand(PROGRAM_SIZE, PIC14_WORDS2BYTES(easy.writesize) + 1);
		if (rc < 0)
			goto done;
		for (uint32_t iy = 0; iy < PIC14_WORDS2BYTES(easy.writesize); iy += 2) {
			assert(ix < easy.startaddr);
			flash[ix++] = ((easy.buffer[iy]) | (easy.buffer[iy + 1] << 8));
		}
	}
	for (uint32_t i = 0; i < easy.startaddr; i += 8) {
		if (pic_isBlank14(&flash[i], 8))
			continue;
		fprintf(stdout, "[%04X]", i);
		for (uint32_t j = 0; j < 8; ++j)
			fprintf(stdout, " %04X", flash[i + j]);
		fputc('\n', stdout);
		++c;
	}
	if (!c)
		fprintf(stdout, "[EMPTY]\n");
done:
	xfree(flash);

	return 0;
}

/*******************************************************************************
 *
 * Dump device EEPROM
 *
 ******************************************************************************/
int
pic14_dumpEEPROM(void)
{
	int rc = 0, c = 0;
	uint8_t *eeprom = xmalloc(easy.eesize * sizeof(uint8_t), -1);
	program_t *program = (program_t *)&easy.buffer[0]; /* COMMAND */

	for (uint32_t address = 0; address < easy.eesize; ++address) {
		program->command = COMMAND_EE_READ;
		program->addru = 0;
		program->addrh = (address & 0x00FF00) >> 8;
		program->addrl = (address & 0x0000FF);
		program->datasize = 1; /* CHECKSUM */
		programChecksum(0);
		rc = doCommand(PROGRAM_SIZE, 2);
		if (rc < 0)
			goto done;
		eeprom[address] = easy.buffer[0];
	}
	for (uint32_t i = 0; i < easy.eesize; i += 16) {
		if (pic_isBlank8(&eeprom[i], 16))
			continue;
		fprintf(stdout, "[%04X]", i);
		for (uint32_t j = 0; j < 16; ++j)
			fprintf(stdout, " %02X", eeprom[i + j]);
		fputc('\n', stdout);
		++c;
	}
	if (!c)
		fprintf(stdout, "[EMPTY]\n");
done:
	xfree(eeprom);

	return 0;
}

/*******************************************************************************
 *
 * Read HEX file and upload to FLASH/EEPROM
 *
 ******************************************************************************/
int
pic14_program(char *filename)
{
	uint32_t nflash = 0, neeprom = 0;
	uint16_t *flash = xmalloc((easy.startaddr << 1) * sizeof(uint16_t), -1);
	uint16_t *eeprom = xmalloc((easy.eesize << 1) * sizeof(uint16_t), -1);

	int rc = readFile(filename, flash, eeprom, &nflash, &neeprom);
	assert(nflash <= (easy.startaddr << 1));
	assert(neeprom <= (easy.eesize << 1));
	if (rc < 0) {
		fprintf(stderr, "Failed to open hex file [%s].\n", filename);
		goto done;
	}
	if (nflash == 0 && neeprom == 0) {
		fprintf(stderr, "No data in hex file [%s].\n", filename);
		--rc;
		goto done;
	}
	if (nflash) {
		pic14_fixGOTO(flash);
		rc = pic14_uploadFlash(flash);
		if (rc < 0)
			goto done;
	}
	if (neeprom) {
		rc = pic14_uploadEEPROM(eeprom);
	}
done:
	xfree(eeprom);
	xfree(flash);

	return rc;
}
