/*------------------------------------------------------------------------------
;
;   Easy Loader.
;
;   Compatible with The Cowboy Boot Loader for Enhanced PIC12/PIC16,
;                   The Wellington Boot Loader for PIC18 and
;                   The Oxford Boot Loader for dsPIC/PIC24.
;
;   Copyright (c) 2014 - 2017 Darron M Broad
;   Copyright (c) 2014 Gerhard Bertelsmann
;
;-------------------------------------------------------------------------------

;-------------------------------------------------------------------------------
;   This file is part of Easy Loader.
;
;   Easy Loader is free software: you can redistribute it and/or
;   modify it under the terms of the GNU General Public License as published
;   by the Free Software Foundation.
;
;   Easy Loader is distributed in the hope that it will be
;   useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
;   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
;   GNU General Public License for more details.
;
;   You should have received a copy of the GNU General Public License along
;   with Easy Loader. If not, see http://www.gnu.org/licenses/
;-----------------------------------------------------------------------------*/

#include "easy-loader.h"

/******************************************************************************
 *
 * Session
 *
 *****************************************************************************/

extern session_t easy;

/*******************************************************************************
 *
 * Fix GOTO
 *
 *  Preserve boot loader reset vector and store application reset vector in
 *  bootloader GOTO application vector.
 *
 ******************************************************************************/
static void
pic24_fixGOTO(uint16_t *flash)
{
	if (!pic_isBlank16(flash, easy.erasesize << 2)) {
		uint32_t gotoapp = (easy.startaddr << 1) - 8;

		/* GOTO application */
		flash[gotoapp + 0] = flash[0] & 0xFF;
		flash[gotoapp + 1] = flash[1] & 0xFF;
		flash[gotoapp + 2] = flash[2] & 0xFF;
		flash[gotoapp + 4] = flash[4] & 0xFF;
		flash[gotoapp + 5] = flash[5] & 0xFF;
		flash[gotoapp + 6] = flash[6] & 0xFF;

		/* GOTO bootloader */
		flash[0] = 0xFF & easy.startaddr;
		flash[1] = 0xFF & (easy.startaddr >> 8);
		flash[2] = 0x04;
		flash[4] = 0xFF & (easy.startaddr >> 16);
		flash[5] = 0;
		flash[6] = 0;
	}
}

/*******************************************************************************
 *
 * Memory copy and compare
 *
 ******************************************************************************/
static void
pic24_memcpy(uint8_t *dest, uint16_t *flash)
{
	if (easy.option & SRCPACKED) {
		for (uint32_t i = 0; i < (easy.writesize << 2); i += 8) {
			*dest++ = flash[i + 0]; /* Low   byte */
			*dest++ = flash[i + 1]; /* High  byte */
			*dest++ = flash[i + 2]; /* Upper byte */
			*dest++ = flash[i + 6]; /* Upper byte */
			*dest++ = flash[i + 4]; /* Low   byte */
			*dest++ = flash[i + 5]; /* High  byte */
		}
	} else {
		for (uint32_t i = 0; i < (easy.writesize << 2); i += 4) {
			*dest++ = flash[i + 2]; /* Upper byte */
			*dest++ = flash[i + 0]; /* Low   byte */
			*dest++ = flash[i + 1]; /* High  byte */
		}
	}
}

static int
pic24_memcmp(uint16_t *s1, uint8_t *s2, uint32_t n)
{
	uint32_t i, j = 0, wordout, wordin;

	for (i = 0; i < n; i += 3, j += 4) {
		wordin = (s2[i] << 16) |	/* Upper byte */
			(s2[i + 1]) |		/* Low   byte */
			(s2[i + 2] << 8);	/* High  byte */
		wordout = (s1[j] & 0xFF) |
			((s1[j + 1] & 0xFF) << 8) |
			((s1[j + 2] & 0xFF) << 16);
		if (wordin != wordout)
			return -1;
	}

	return 0;
}

/*******************************************************************************
 *
 * Erase row
 *
 ******************************************************************************/
int
pic24_erase_row(uint32_t raddress)
{
	program_t *program = (program_t *)&easy.buffer[0]; /* COMMAND */

	/* Erase row */
	program->command = COMMAND_FLASH_ERASE;
	program->addru = (raddress & 0xFF0000) >> 16;
	program->addrh = (raddress & 0x00FF00) >> 8;
	program->addrl = (raddress & 0x0000FF);
	program->datasize = 1; /* CHECKSUM */
	programChecksum(0);
	if (doCommand(PROGRAM_SIZE, 1) < 0)
		return -1;

	return 0;
}

/*******************************************************************************
 *
 * Verify row
 *
 ******************************************************************************/
static int
pic24_verify_row(uint16_t *flash, uint32_t address)
{
	program_t *program = (program_t *)&easy.buffer[0]; /* COMMAND */
	uint32_t faddress = (address << 1);

	program->command = COMMAND_FLASH_READ;
	program->addru = (address & 0xFF0000) >> 16;
	program->addrh = (address & 0x00FF00) >> 8;
	program->addrl = (address & 0x0000FF);
	program->datasize = 1; /* CHECKSUM */
	programChecksum(0);
	if (doCommand(PROGRAM_SIZE, PIC24_WORDS2BYTES(easy.writesize) + 1) < 0)
		return -1;
	if (pic24_memcmp(&flash[faddress], easy.buffer, PIC24_WORDS2BYTES(easy.writesize)) == 0)
		fprintf(stdout, " VERIFY OK\n");
	else
		fprintf(stderr, " VERIFY ERROR\n");

	return 0;
}

/*******************************************************************************
 *
 * Upload to flash
 *
 *  Erase and Write rows. For some devices the erase row is larger than
 *  the write row.
 *
 ******************************************************************************/
static int
pic24_uploadFlash(uint16_t *flash)
{
	program_t *program = (program_t *)&easy.buffer[0]; /* COMMAND */

	/*
	 * For each erase row
	 */

	for (uint32_t eaddress = 0; eaddress < easy.startaddr; eaddress += (easy.erasesize << 1)) {
		uint32_t faddress = (eaddress << 1);
		assert(faddress < (easy.startaddr << 1));
		if (pic_isBlank16(&flash[faddress], easy.erasesize << 2))
			continue;

		/*
		 * First write row
		 */

		if (pic_isBlank16(&flash[faddress], easy.writesize << 2)) {

			/*
			 * Erase only
			 */

			/* Simulate? */
			if (!easy.simulate) {
				/* Erase row */
				if (pic24_erase_row(eaddress) < 0)
					return -1;
			}
			/* Information */
			if (easy.info)
				fprintf(easy.info, "ERASE FLASH 0x%06X\n", eaddress);
		} else {

			/*
			 * Erase then write
			 */

			/* Simulate? */
			if (!easy.simulate) {
				/* Erase row then write row */
				program->command = COMMAND_FLASH_ERASE;
				program->addru = (eaddress & 0xFF0000) >> 16;
				program->addrh = (eaddress & 0x00FF00) >> 8;
				program->addrl = (eaddress & 0x0000FF);
				program->datasize = 1 + easy.writesize; /* CHECKSUM + data */
				pic24_memcpy(program->data, &flash[faddress]);
				programChecksum(PIC24_WORDS2BYTES(easy.writesize));
				if (doCommand(PROGRAM_SIZE + PIC24_WORDS2BYTES(easy.writesize), 1) < 0)
					return -1;
			}
			/* Information */
			if (easy.info)
				fprintf(easy.info, "ERASE FLASH 0x%06X %4d BYTES\n",
					eaddress, PIC24_WORDS2BYTES(easy.writesize));
			/* Verify */
			if (easy.verify) {
				if (pic24_verify_row(flash, eaddress) < 0)
					return -1;
			}
		}

		/*
		 * Next write row(s)
		 */

		for (uint32_t i = easy.writesize; i < easy.erasesize; i += easy.writesize) {
			uint32_t raddress = eaddress + (i << 1);
			faddress = (raddress << 1);
			assert(faddress < (easy.startaddr << 1));
			if (pic_isBlank16(&flash[faddress], easy.writesize << 2))
				continue;

			/*
			 * Write
			 */

			/* Simulate? */
			if (!easy.simulate) {
				/* Write row */
				program->command = COMMAND_FLASH_WRITE;
				program->addru = (raddress & 0xFF0000) >> 16;
				program->addrh = (raddress & 0x00FF00) >> 8;
				program->addrl = (raddress & 0x0000FF);
				program->datasize = 1 + easy.writesize; /* CHECKSUM + data */
				pic24_memcpy(program->data, &flash[faddress]);
				programChecksum(PIC24_WORDS2BYTES(easy.writesize));
				if (doCommand(PROGRAM_SIZE + PIC24_WORDS2BYTES(easy.writesize), 1) < 0)
					return -1;
			}
			/* Information */
			if (easy.info)
				fprintf(easy.info, "WRITE FLASH 0x%06X %4d BYTES\n",
					raddress, PIC24_WORDS2BYTES(easy.writesize));
			/* Verify */
			if (easy.verify) {
				if (pic24_verify_row(flash, raddress) < 0)
					return -1;
			}
		}
	}

	return 0;
}

/*******************************************************************************
 *
 * Upload to EEPROM
 *
 ******************************************************************************/
static int
pic24_uploadEEPROM(uint16_t *eeprom)
{
	program_t *program = (program_t *)&easy.buffer[0]; /* COMMAND */

	for (uint32_t i = 0; i < (easy.eesize << 1); i += 4) {
		uint32_t eaddress = (0x800000 - easy.eesize) + (i >> 1);

		if (eeprom[i] == 0xFFFF && eeprom[i + 1] == 0xFFFF)
			continue;

		/* Simulate? */
		if (!easy.simulate) {
			/* Write word */
			program->command = COMMAND_EE_WRITE;
			program->addru = (eaddress & 0xFF0000) >> 16;
			program->addrh = (eaddress & 0x00FF00) >> 8;
			program->addrl = (eaddress & 0x0000FF);
			program->datasize = 2; /* CHECKSUM + data */
			program->data[0] = eeprom[i];		/* Low   byte */
			program->data[1] = eeprom[i + 1];	/* High  byte */
			program->data[2] = 0;			/* Upper byte */
			programChecksum(PIC24_WORDS2BYTES(1));
			if (doCommand(PROGRAM_SIZE + PIC24_WORDS2BYTES(1), 1) < 0)
				return -1;
		}
		/* Information */
		if (easy.info)
			fprintf(easy.info, "WRITE EEPROM 0x%04X = 0x%04X\n",
				eaddress, eeprom[i] | eeprom[i + 1] << 8);
		/* Verify */
		if (easy.verify) {
			printf("%s() UNIMPLEMENTED\n", __func__);
		}
	}

	return 0;
}

/*******************************************************************************
 *
 * Dump device flash
 *
 ******************************************************************************/
int
pic24_dumpFlash(void)
{
	int rc = 0, c = 0;
	program_t *program = (program_t *)&easy.buffer[0];
	uint32_t asize = easy.startaddr >> 1, *flash = xmalloc(asize * sizeof(uint32_t), -1), ix = 0;

	for (uint32_t address = 0; address < easy.startaddr; address += (easy.writesize << 1)) {
		program->command = COMMAND_FLASH_READ;
		program->addru = (address & 0xFF0000) >> 16;
		program->addrh = (address & 0x00FF00) >> 8;
		program->addrl = (address & 0x0000FF);
		program->datasize = 1; /* CHECKSUM */
		programChecksum(0);
		rc = doCommand(PROGRAM_SIZE, PIC24_WORDS2BYTES(easy.writesize) + 1);
		if (rc < 0)
			goto done;
		for (uint32_t iy = 0; iy < PIC24_WORDS2BYTES(easy.writesize); iy += 3) {
			assert(ix < asize);
			flash[ix++] = easy.buffer[iy] << 16 |	/* Upper byte */
				easy.buffer[iy + 1] |		/* Low   byte */
				easy.buffer[iy + 2] << 8;	/* High  byte */
		}
	}
	for (uint32_t i = 0; i < asize; i += 8) {
		if (pic_isBlank24(&flash[i], 8))
			continue;
		fprintf(stdout, "[%06X]", i << 1);
		for (uint32_t j = 0; j < 8; ++j)
			fprintf(stdout, " %06X", flash[i + j]);
		fputc('\n', stdout);
		++c;
	}
	if (!c)
		fprintf(stdout, "[EMPTY]\n");
done:
	xfree(flash);

	return rc;
}

/*******************************************************************************
 *
 * Dump device EEPROM
 *
 ******************************************************************************/
int
pic24_dumpEEPROM(void)
{
	int rc = 0, c = 0;
	program_t *program = (program_t *)&easy.buffer[0];
	uint16_t *eeprom = xmalloc((easy.eesize >> 1) * sizeof(uint16_t), -1), ix = 0;
	uint32_t eeaddr = (0x800000 - easy.eesize);

	for (uint32_t address = eeaddr; address < 0x800000; address += 2) {
		program->command = COMMAND_EE_READ;
		program->addru = (address & 0xFF0000) >> 16;
		program->addrh = (address & 0x00FF00) >> 8;
		program->addrl = (address & 0x0000FF);
		program->datasize = 1; /* CHECKSUM */
		programChecksum(0);
		rc = doCommand(PROGRAM_SIZE, 2 + 1);
		if (rc < 0)
			goto done;
		eeprom[ix++] = easy.buffer[0] | easy.buffer[1] << 8;
	}
	for (uint32_t i = 0; i < (easy.eesize >> 1); i += 8) {
		if (pic_isBlank16(&eeprom[i], 8))
			continue;
		fprintf(stdout, "[%06X]", eeaddr + (i << 1));
		for (uint32_t j = 0; j < 8; ++j)
			fprintf(stdout, " %04X", eeprom[i + j]);
		fputc('\n', stdout);
		++c;
	}
	if (!c)
		fprintf(stdout, "[EMPTY]\n");
done:
	xfree(eeprom);

	return rc;
}

/*******************************************************************************
 *
 * Read HEX file and upload to FLASH/EEPROM
 *
 ******************************************************************************/
int
pic24_program(char *filename)
{
	uint32_t nflash = 0, neeprom = 0;
	uint16_t *flash = xmalloc((easy.startaddr << 1) * sizeof(uint16_t), -1);
	uint16_t *eeprom = xmalloc((easy.eesize << 1) * sizeof(uint16_t), -1);

	int rc = readFile(filename, flash, eeprom, &nflash, &neeprom);
	assert(nflash <= (easy.startaddr << 1));
	assert(neeprom <= (easy.eesize << 1));
	if (rc < 0) {
		fprintf(stderr, "Failed to open hex file [%s].\n", filename);
		goto done;
	}
	if (nflash == 0 && neeprom == 0) {
		fprintf(stderr, "No data in hex file [%s].\n", filename);
		--rc;
		goto done;
	}
	if (nflash) {
		pic24_fixGOTO(flash);
		rc = pic24_uploadFlash(flash);
		if (rc < 0)
			goto done;
	}
	if (neeprom) {
		rc = pic24_uploadEEPROM(eeprom);
	}
done:
	xfree(eeprom);
	xfree(flash);

	return rc;
}
