/*------------------------------------------------------------------------------
;
;   Easy Loader.
;
;   Compatible with The Cowboy Boot Loader for Enhanced PIC12/PIC16,
;                   The Wellington Boot Loader for PIC18 and
;                   The Oxford Boot Loader for dsPIC/PIC24.
;
;   Copyright (c) 2014 - 2017 Darron M Broad
;   Copyright (c) 2014 Gerhard Bertelsmann
;
;-------------------------------------------------------------------------------

;-------------------------------------------------------------------------------
;   This file is part of Easy Loader.
;
;   Easy Loader is free software: you can redistribute it and/or
;   modify it under the terms of the GNU General Public License as published
;   by the Free Software Foundation.
;
;   Easy Loader is distributed in the hope that it will be
;   useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
;   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
;   GNU General Public License for more details.
;
;   You should have received a copy of the GNU General Public License along
;   with Easy Loader. If not, see http://www.gnu.org/licenses/
;-----------------------------------------------------------------------------*/

#include "easy-loader.h"

/******************************************************************************
 *
 * Session
 *
 *****************************************************************************/

extern session_t easy;

/*******************************************************************************
 *
 * Return `speed_t' for Given Baud Rate
 *
 ******************************************************************************/
static speed_t
serial_speed(uint32_t baudrate)
{
	static baudrate_t rates[] = {
	{0, B0},
	{50, B50},
	{75, B75},
	{110, B110},
	{134, B134},
	{150, B150},
	{200, B200},
	{300, B300},
	{600, B600},
	{1200, B1200},
	{1800, B1800},
	{2400, B2400},
	{4800, B4800},
#ifdef B7200
	{7200, B7200},
#endif
	{9600, B9600},
#ifdef B14400
	{14400, B14400},
#endif
#ifdef B19200
	{19200, B19200},
#else
	{19200, EXTA},
#endif
#ifdef B28800
	{28800, B28800},
#endif
#ifdef B38400
	{38400, B38400},
#else
	{38400, EXTB},
#endif
#ifdef B57600
	{57600, B57600},
#endif
#ifdef B76800
	{76800, B76800},
#endif
#ifdef B115200
	{115200, B115200},
#endif
#ifdef B230400
	{230400, B230400},
#endif
#ifdef B460800
	{460800, B460800},
#endif
#ifdef B500000
	{500000, B500000},
#endif
#ifdef B576000
	{576000, B576000},
#endif
#ifdef B921600
	{921600, B921600},
#endif
#ifdef B1000000
	{1000000, B1000000},
#endif
#ifdef B1152000
	{1152000, B1152000},
#endif
#ifdef B1500000
	{1500000, B1500000},
#endif
#ifdef BHANGUP0
	{HANGUP0, BHANGUP0},
#endif
#ifdef B2500000
	{2500000, B2500000},
#endif
#ifdef B3000000
	{3000000, B3000000},
#endif
#ifdef B3500000
	{3500000, B3500000},
#endif
#ifdef B4000000
	{4000000, B4000000},
#endif
	{UINT32_MAX, B9600},
	};
	int i = 0;

	while (baudrate > rates[i++].baud)
		;

	return rates[--i].speed;
}

/*******************************************************************************
 *
 * Open serial device
 *
 *  Hang up RTS/DTR If Required
 *
 ******************************************************************************/
int
openDevice(const char *dev)
{
	int fd, status;
	struct termios options;
	speed_t speed;

	fd = open(dev, O_RDWR | O_NOCTTY | O_NONBLOCK);
	if (fd < 0) {
		return fd;
	}
	if (fcntl(fd, F_SETFL, O_NONBLOCK | fcntl(fd, F_GETFL)) < 0) {
		close(fd);
		return -1;
	}
	if (tcgetattr(fd, &options) < 0) {
		close(fd);
		return -1;
	}

	/*
	 * Raw mode 8N2
	 *
	 *  Linux TERMIOS(3)
	 */

	options.c_iflag &= ~(IGNBRK | BRKINT | PARMRK | ISTRIP | INLCR | IGNCR | ICRNL | IXON);
	options.c_oflag &= ~(OPOST);
	options.c_cflag &= ~(CSIZE | PARENB);
	options.c_cflag |= (CS8 | CSTOPB);
	options.c_lflag &= ~(ECHO | ECHONL | ICANON | ISIG | IEXTEN);

	options.c_cc[VMIN] = 0;
	options.c_cc[VTIME] = 0;

	speed = serial_speed(easy.baudrate);

	if (cfsetispeed(&options, speed) < 0) {
		close(fd);
		return -1;
	}
	if (cfsetospeed(&options, speed) < 0) {
		close(fd);
		return -1;
	}
	if (tcsetattr(fd, TCSANOW, &options) < 0) {
		close(fd);
		return -1;
	}
	if (tcflush(fd, TCIOFLUSH) < 0) {
		close(fd);
		return -1;
	}

	/*
	 * Ensure RTS/DTR are asserted
	 *
	 *  Assert RTS/DTR to put a device into reset
	 */

	ioctl(fd, TIOCMGET, &status);
	status = status | TIOCM_RTS | TIOCM_DTR;
	ioctl(fd, TIOCMSET, &status); /* LOGIC 0 */

	/*
	 * Send BREAK
	 */

	if (easy.brk) {
		ioctl(fd, TIOCSBRK, 0); 
		usleep(easy.brk);
		ioctl(fd, TIOCCBRK, 0); 
	}
	
	/*
	 * Send BREAK to reset controller, may de-assert RTS
	 *
	 *  Winklepicker modes 1 or 2
	 */
	if (easy.winkle) {
		ioctl(fd, TIOCSBRK, 0); 
		usleep(55000); /* 55ms */
		ioctl(fd, TIOCCBRK, 0); 

	 	/*
		 * Winklepicker mode 2 reset
		 */
		if (easy.winkle == 2) {
			status &= ~TIOCM_RTS;
			ioctl(fd, TIOCMSET, &status); /* LOGIC 1 */
			usleep(5000); /* 5ms */
			status |= TIOCM_RTS;
			ioctl(fd, TIOCMSET, &status); /* LOGIC 0 */
			usleep(10000); /* 10ms */
			status &= ~TIOCM_RTS;
			ioctl(fd, TIOCMSET, &status); /* LOGIC 1 */
		}
	}

	/*
	 * Arduino reset (DTR + CAP)
	 */

	if (easy.wiener) {
		usleep(easy.wiener);
		status &= ~TIOCM_DTR;
		ioctl(fd, TIOCMSET, &status); /* LOGIC 1 */
		usleep(easy.wiener);
		status |= TIOCM_DTR;
		ioctl(fd, TIOCMSET, &status); /* LOGIC 0 */
		usleep(easy.wiener);
		status &= ~TIOCM_DTR;
		ioctl(fd, TIOCMSET, &status); /* LOGIC 1 */
	}

	/*
	 * RTS/DTR hang up
	 *
	 *  De-assert RTS/DTR to bring a device out of reset
	 */

	if (easy.rts) {
		usleep(easy.rts);
		status &= ~TIOCM_RTS;
		ioctl(fd, TIOCMSET, &status); /* LOGIC 1 */
	}

	if (easy.dtr) {
		usleep(easy.dtr);
		status &= ~TIOCM_DTR;
		ioctl(fd, TIOCMSET, &status); /* LOGIC 1 */
	}

	/*
	 * Boot loader start up delay after reset
	 */

	usleep(easy.delay);

	return fd;
}

void
closeDevice(void)
{
	/*
	 * Send BREAK to reset controller with RTS asserted
	 *
	 *  Winklepicker mode 2 only
	 */
	if (easy.winkle == 2) {
		ioctl(easy.fd, TIOCSBRK, 0); 
		usleep(55000); /* 55ms */
		ioctl(easy.fd, TIOCCBRK, 0); 
	}
#if 0
	tcflush(easy.fd, TCIOFLUSH);
#endif
}
